@@ -83,9 +83,8 @@ static void dev_usage_to_text(struct printbuf *out,
 
 		prt_printf(out, "\r%llu\r", u->d[type].buckets);
 
-		u64 fragmented = u->d[type].buckets * u->bucket_size - sectors;
-		if (fragmented)
-			prt_units_u64(out, fragmented << 9);
+		if (u->d[type].fragmented)
+			prt_units_u64(out, u->d[type].fragmented << 9);
 		prt_printf(out, "\r\n");
 	}
 
@@ -116,9 +115,7 @@ static int dev_by_label_cmp(const void *_l, const void *_r)
 
 static u64 dev_leaving_dev(struct bch_ioctl_query_accounting *in, unsigned dev)
 {
-	for (struct bkey_i_accounting *a = in->accounting;
-	     a < (struct bkey_i_accounting *) ((u64 *) in->accounting + in->accounting_u64s);
-	     a = bkey_i_to_accounting(bkey_next(&a->k_i))) {
+	for_each_accounting(in, a) {
 		struct disk_accounting_pos acc_k;
 		bpos_to_disk_accounting_pos(&acc_k, a->k.p);
 
@@ -305,9 +302,7 @@ static int accounting_p_cmp(const void *_l, const void *_r)
 static void accounting_sort(darray_accounting_p *sorted,
 			    struct bch_ioctl_query_accounting *in)
 {
-	for (struct bkey_i_accounting *a = in->accounting;
-	     a < (struct bkey_i_accounting *) ((u64 *) in->accounting + in->accounting_u64s);
-	     a = bkey_i_to_accounting(bkey_next(&a->k_i)))
+	for_each_accounting(in, a)
 		if (darray_push(sorted, a))
 			die("memory allocation failure");
 
@@ -320,9 +315,7 @@ static void accounting_swab_if_old(struct bch_ioctl_query_accounting *in)
 
 	if (kernel_version &&
 	    kernel_version < bcachefs_metadata_version_disk_accounting_big_endian)
-		for (struct bkey_i_accounting *a = in->accounting;
-		     a < (struct bkey_i_accounting *) ((u64 *) in->accounting + in->accounting_u64s);
-		     a = bkey_i_to_accounting(bkey_next(&a->k_i)))
+		for_each_accounting(in, a)
 			bch2_bpos_swab(&a->k.p);
 }
 
@@ -684,6 +677,7 @@ static void fs_usage_usage(void)
 	     "  -H, --help                   Display this help and exit\n"
 	     "\n"
 	     "Report bugs to <linux-bcachefs@vger.kernel.org>");
+	exit(EXIT_SUCCESS);
 }
 
 int cmd_fs_usage(int argc, char *argv[])
@@ -745,14 +739,14 @@ int cmd_fs_usage(int argc, char *argv[])
 int fs_usage(void)
 {
 	puts("bcachefs fs - manage a running filesystem\n"
-	     "Usage: bcachefs fs <CMD> [OPTIONS]\n"
+	     "Usage: bcachefs fs <usage|top> [OPTION]...\n"
 	     "\n"
 	     "Commands:\n"
 	     "  usage                        Display detailed filesystem usage\n"
 	     "  top                          Show runtime performance information\n"
 	     "\n"
 	     "Report bugs to <linux-bcachefs@vger.kernel.org>");
-	return 0;
+	exit(EXIT_SUCCESS);
 }
 
 int fs_cmds(int argc, char *argv[])
