@@ -11,6 +11,7 @@
 
 #include "alloc/backpointers.h"
 #include "alloc/buckets_waiting_for_journal.h"
+#include "alloc/discard.h"
 #include "alloc/disk_groups.h"
 #include "alloc/foreground.h"
 #include "alloc/replicas.h"
@@ -27,11 +28,12 @@
 #include "data/checksum.h"
 #include "data/compress.h"
 #include "data/copygc.h"
-#include "data/ec.h"
+#include "data/ec/create.h"
+#include "data/ec/init.h"
 #include "data/move.h"
 #include "data/nocow_locking.h"
 #include "data/read.h"
-#include "data/reconcile.h"
+#include "data/reconcile/work.h"
 #include "data/write.h"
 
 #ifdef CONFIG_DEBUG_FS
