@@ -14,4 +14,7 @@ int bch2_scan_device_sbs(char *, sb_names *ret);
 
 char *bch2_scan_devices(char *);
 
+/* src/http.rs: */
+void start_http(const char *listen);
+
 #endif /* _BCACHEFS_TOOLS_RUST_TO_C_H */
